#!/usr/bin/env python

# codecs2.py - encoding transparent access to file objects.

# Copyright (c) 2007 Masaaki Shibata <mshibata@emptypage.jp>
# 
# Permission is hereby granted, free of charge, to any person 
# obtaining a copy of this software and associated documentation files 
# (the "Software"), to deal in the Software without restriction, 
# including without limitation the rights to use, copy, modify, merge, 
# publish, distribute, sublicense, and/or sell copies of the Software, 
# and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be 
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS 
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN 
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
# SOFTWARE.


import os
import sys
import codecs

try:
    from msvcrt import setmode as _setmode
    _setbin = lambda fd: _setmode(fd, os.O_BINARY)
except ImportError:
    _setbin = lambda fd: 0


__all__ = ['StreamReaderWriter', 'interface', 'open',
           'stdin', 'stdout', 'stderr', 'bstdin', 'bstdout', 'bstderr']
__version__ = '0.1.0'

class StreamReaderWriter(codecs.StreamReaderWriter):
    """Subclasses codecs.StreamReaderWriter to convert newline 
    characters correctly.
    
    Note: Current (2.5) implementation of codecs.StreamReaderWriter 
    doesn't provide 'text mode' newline character conversion. This 
    complements the conversion.
    """
    
    def __init__(self, stream, Reader, Writer, errors, linesep=os.linesep):
        """Creates new codecs2.StreamReaderWriter object.
        
        stream, Reader, Writer, errors
            Same as its baseclass. See codecs.StreamReaderWriter.
        linesep
            Newline character, '\\r', '\\n', or '\\r\\n'. Default is 
            OS dependent.
        """
        
        codecs.StreamReaderWriter.__init__(self,
                                           stream, Reader, Writer, errors)
        self.linesep = linesep
    
    
    def read(self, size=-1):
        return self.reader.read(size).replace(self.linesep, '\n')
    
    def readline(self, size=None):
        return self.reader.readline(size).replace(self.linesep, '\n')

    def readlines(self, sizehint=None):
        return [line.replace(self.linesep, '\n') for line
                in self.reader.readlines(sizehint)]

    def next(self):
        return self.reader.next().replace(self.linesep, '\n')

    def write(self, data):
        return self.writer.write(data.replace('\n', self.linesep))

    def writelines(self, list):
        return self.writer.writelines([line.replace('\n', self.linesep)
                                       for line in list])


def _dupbin(fp):
    fd = os.dup(fp.fileno())
    _setbin(fd)
    return os.fdopen(fd, fp.mode)


def interface(fp, encoding, errors='strict', linesep=os.linesep):
    """Provides encoding-transparent access interface to file (like) object.
    
    Note: This is similar to codecs.EncodedFile but this can also 
    treats newline separators correctly.
    """
    reader, writer = codecs.lookup(encoding)[2:]
    return StreamReaderWriter(fp, reader, writer, errors, linesep)


def open(filename, mode, encoding=None, errors='strict', bufsize=1,
         linesep=os.linesep):
    """Creates file object with encoding-transparent access interfaces.
    
    Note: This is similar to codecs.open but this can also treats 
    newline separators correctly.
    """
    if not encoding:
        fp = file(filename, mode, bufsize)
    else:
        if 'b' not in mode:
            mode += 'b'
        fp = interface(file(filename, mode, bufsize),
                       encoding, errors, linesep)
        fp.encoding = encoding
    return fp


def stdin(encoding=None, errors='strict', linesep=os.linesep):
    """Provides encoding-transparent access interface to stdin.
    """
    if not encoding:
        encoding = sys.stdin.encoding
    return interface(_dupbin(sys.stdin), encoding, errors, linesep)


def stdout(encoding=None, errors='strict', linesep=os.linesep):
    """Provides encoding-transparent access interface to stdout.
    """
    if not encoding:
        encoding = sys.stdout.encoding
    return interface(_dupbin(sys.stdout), encoding, errors, linesep)


def stderr(encoding=None, errors='strict', linesep=os.linesep):
    """Provides encoding-transparent access interface to stderr.
    """
    if not encoding:
        encoding = sys.stderr.encoding
    return interface(_dupbin(sys.stderr), encoding, errors, linesep)




_bstdin = _dupbin(sys.stdin)
def bstdin():
    """Provides binary mode stdin.
    """
    return _bstdin


_bstdout = _dupbin(sys.stdout)
def bstdout():
    """Provides binary mode stdout.
    """
    return _bstdout


_bstderr = _dupbin(sys.stderr)
def bstderr():
    """Provides binary mode stderr.
    """
    return _bstderr


_usage = """codecs2.py [options] <filenames...>
options:
  -e | --encoding-in <encoding>     input encoding
  -n | --encoding-out <encoding>    output encoding
  -w | --newline-in <newline>       input newline code - CR | LF | CRLF
  -l | --newline-out <newline>      output newline code - CR | LF | CRLF
  -r | --replace                    replace original file (be careful!)
  -h | --help                       show this help
"""


def main():
    import getopt
    import locale
    import tempfile
    
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'e:n:w:l:rvh',
                                   ['encoding-in=', 'encoding-out=',
                                    'newline-in=', 'newline-out=',
                                    'replace', 'verbose', 'help'])
    except getopt.GetoptError:
        print _usage
        sys.exit(2)
    
    enc_in = enc_out = locale.getpreferredencoding()
    nl_in = nl_out = os.linesep
    replace = False
    verbose = False
    
    for k, v in opts:
        if k in ('-e', '--encoding-in'):
            enc_in = v
        if k in ('-n', '--encoding-out'):
            enc_out = v
        if k in ('-w', '--newline-in'):
            nl_in = {'cr': '\r', 'lf': '\n', 'crlf': '\r\n'}[v.lower()]
        if k in ('-l', '--newline-out'):
            nl_out = {'cr': '\r', 'lf': '\n', 'crlf': '\r\n'}[v.lower()]
        if k in ('-r', '--replace'):
            replace = True
        if k in ('-v', '--verbose'):
            verbose = True
        if k in ('-h', '--help'):
            print _usage
            sys.exit()
    
    if not sys.stdin.isatty():
        args.insert(0, stdin(enc_in, 'replace', nl_in))
    if not args:
        print _usage
        sys.exit(2)
    
    if verbose:
        nlname = {'\r': 'CR', '\n': 'LF', '\r\n': 'CRLF'}
        print >>sys.stderr, '%s (%s) --> %s (%s)' \
        % (enc_in, nlname[nl_in], enc_out, nlname[nl_out])
    
    def output(fp_in, fp_out):
        for line in fp_in:
            fp_out.write(line)
    
    fp_stdout = stdout(enc_out, 'replace', nl_out)
    for f in args:
        if not isinstance(f, basestring):
            # f is stdin
            output(f, fp_stdout)
            continue
        # else
        fp_in = open(f, 'r', enc_in, 'replace', linesep=nl_in)
        if replace:
            if verbose:
                print >>sys.stderr, 'converting: \'%s\'' % f
            fd2, filename2 = tempfile.mkstemp(dir=os.path.dirname(f))
            fp_out = interface(os.fdopen(fd2, 'wb'),
                               enc_out, 'replace', nl_out)
            output(fp_in, fp_out)
            fp_in.close()
            fp_out.close()
            os.remove(f)
            os.rename(filename2, f)
        else:
            output(fp_in, fp_stdout)


if __name__ == '__main__':
    main()


